from datetime import datetime
import json
from pathlib import Path
import pandas as pd
import requests
import keyring
from requests.auth import HTTPBasicAuth
from dataclasses import dataclass
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed
# Egne klasser
from .cvr_response_class import CVRResponse
from .cvr_regnskab_class import Regnskab
from .cvr_response_owner_class import CVROwner

@dataclass
class CVRAPI:
    """Klasse som trækker data fra cvr registret. Det er muligt at hente cvr data (antal ansatte, branche etc.)
    og regnskabsdata.
    
    Adgangen til CVRs api antager at credentials er tilføjet windows credentials. Se https://support.microsoft.com/en-us/windows/accessing-credential-manager-1b5c916a-6a16-889f-8581-fc16e8165ac0
    Det antages at de tiføjede credentials er navngivet cvr_api
    """
    
    def __post_init__(self):
        self._url = self._url_virk
        self.__test_setup()
        
    def __test_setup(self):
        self.__creds = keyring.get_credential("cvr_api", None)
        if self.__creds is None:
            raise ValueError("Du skal tilføje brugernavn og password til windows credentials kaldet cvr_api.")
        test_query = {"query": {"bool": {"must": [{"term": {"Vrvirksomhed.cvrNummer": "29035695"}}]}}}
        response = self._response(test_query)
        response.raise_for_status()
        print("Forbindelse til CVR etableret korrekt!")
        
    @property
    def _headers(self) -> dict:
        return {"content-type": "application/json", "cache-control": "no-cache"}
    
    @property
    def _url(self) -> str:
        return self._internal_url
    
    @_url.setter
    def _url(self, url: str):
        self._internal_url = url
    
    @property
    def _url_virk(self) -> str:
        return r"http://distribution.virk.dk/cvr-permanent/virksomhed/_search"

    @property
    def _url_penh(self) -> str:
        return r"http://distribution.virk.dk/cvr-permanent/produktionsenhed/_search"
    
    @property
    def _url_regnskab(self) -> str:
        return r"http://distribution.virk.dk/offentliggoerelser/_search"
    
    def _response(self, payload: dict) -> requests.Response:
        resp = requests.request(
            method="POST",
            url=self._url,
            data=json.dumps(payload),
            headers=self._headers,
            auth=HTTPBasicAuth(self.__creds.username, self.__creds.password)
        )
        resp.raise_for_status()
        return resp
   
    def test_query(self, query: dict, url_virk: bool = True) -> json:
        """Funktion som anvendes til at teste nye queries inden indlejring i klassen"""
        self._url = self._url_virk if url_virk else self._url_penh
        response = self._response(query)
        json_cont = json.loads(response.content)
        print(json_cont)
        return self.__parse_responses(json_cont)
    
    def __parse_responses(self, json_cont: json) -> list:
        results = []
        if json_cont["hits"]["total"] == 0:
            return results
        for hit in json_cont["hits"]["hits"]:
            # Grim, grim, grim håndtering
            if r"/virksomhed/" in self._url:
                if hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["nyesteNavn"] is None:
                    continue
                results.append(self.__parse_response(hit))
            else:
                results.append(self.__parse_response_penhed(hit))
        return results

    def __parse_response_penhed(self, hit: dict) -> dict:
        """
        NB! Samles med __parse_respose til én CVRResponse klasse        
        Trækker relevante informationer ud fra den returnerede json-streng. Bør nok håndteres anderledes, da 
        det er forskelligt hvilke info, som reelt er relevante. 
        Lav om til en "response klasse" som kan genanvendes andre steder."""
        result = {}
        if hit["_source"]["VrproduktionsEnhed"]["produktionsEnhedMetadata"]["nyesteNavn"] is None:
            return result
        result["name"] = hit["_source"]["VrproduktionsEnhed"]["produktionsEnhedMetadata"]["nyesteNavn"]["navn"]
        result["pnr"] = hit["_source"]["VrproduktionsEnhed"]["pNummer"]
        try:
            result["cvr"] = hit["_source"]["VrproduktionsEnhed"]["virksomhedsrelation"][0]["cvrNummer"]
        except IndexError:
            print(self.query_term)
            pass
        result["status"] = hit["_source"]["VrproduktionsEnhed"]["produktionsEnhedMetadata"]["sammensatStatus"]
        result["match_score"] = hit["_score"]
        return result
    
    def __parse_response(self, hit: dict) -> dict:
        """Trækker relevante informationer ud fra den returnerede json-streng. Bør nok håndteres anderledes, da 
        det er forskelligt hvilke info, som reelt er relevante. 
        Lav om til en "response klasse" som kan genanvendes andre steder."""
        result = {}
        if hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["nyesteNavn"] is None:
            return result
        result["name"] = hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["nyesteNavn"]["navn"]
        result["cvr"] = hit["_source"]["Vrvirksomhed"]["cvrNummer"]
        result["status"] = hit["_source"]["Vrvirksomhed"]["virksomhedMetadata"]["sammensatStatus"]
        result["match_score"] = hit["_score"]
        return result
    
    def __search_based_on_cvr(self, cvr: str) -> json:
        query = {"query": {"bool": {"must": [{"term": {"Vrvirksomhed.cvrNummer": cvr}}]}}}
        self._url = self._url_virk
        response = self._response(query)
        response.raise_for_status()
        return json.loads(response.content)

    def __search_based_on_cvrp(self, cvr: str) -> json:
        query = {"query": {"bool": {"must": [{"term": {"VrproduktionsEnhed.pNummer": cvr}}]}}}
        self._url = self._url_penh
        response = self._response(query)
        response.raise_for_status()
        return json.loads(response.content)
    
    def cvr_response_raw(self, cvr: int) -> dict:
        response = self.__search_based_on_cvr(str(cvr))
        print(response)
        return response
    
    def cvr_search(self, cvr: list[int]) -> list[CVRResponse]:
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        if not all(isinstance(c, int) for c in cvr):
            raise ValueError("CVR skal være angivet som int")
        results = []
        for c in cvr:
            hits_from_cvr = self.__search_based_on_cvr(str(c))
            results.append(CVRResponse(hits_from_cvr["hits"]["hits"][0]))
        return results
    
    def cvr_search_owners_to_df(self, cvr: list[int], only_active: bool = True) -> pd.DataFrame:
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        if not all(isinstance(c, int) for c in cvr):
            raise ValueError("CVR skal være angivet som int")
                
        print(f"Henter CVR data (ejere) for {len(cvr)} cvr-numre")
        
        results = []
                
        def process_single_cvr(c):
            api_results = []
            hits_from_cvr = self.__search_based_on_cvr(str(c))
            if hits_from_cvr["hits"]["total"] > 0:
                api_results = CVROwner(hits_from_cvr["hits"]["hits"][0], c).get_data()
                return api_results
            else:
                return {
                    "cvr": c,
                    "fejl": "Opslaget gav ingen hits",
                }

        with ThreadPoolExecutor() as executor:
            future_to_cvr = {executor.submit(process_single_cvr, c): c for c in cvr}

            for future in tqdm(as_completed(future_to_cvr), total=len(cvr)):
                result = future.result()

                if isinstance(result, list):
                    results.extend(result)
                else:
                    results.append(result)

        return pd.DataFrame.from_dict(results) if not only_active else pd.DataFrame.from_dict(results).query(r"aktiv==1")
    
    def cvr_search_to_df(self, cvr: list[int], latest: bool = True, measure: str = None) -> pd.DataFrame:
        """Grimt - skal skrives om. Midlertidig løsning for at inkludere historiske data for cvr-nummeret."""
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        
        print(f"Henter CVR data for {len(cvr)} cvr-numre")
        results = []
        
        def process_single_cvr(c):
            hits_from_cvr = self.__search_based_on_cvr(str(c))
            if hits_from_cvr["hits"]["total"] > 0:
                if latest:
                    api_results = CVRResponse(hits_from_cvr["hits"]["hits"][0]).get_data()
                    api_results["hit_type"] = "cvr"
                    return api_results
                else:
                    api_results = CVRResponse(hits_from_cvr["hits"]["hits"][0]).get_data(all=True, unit=measure)
                    return api_results
            else:
                hits_pnr = self.__search_based_on_cvrp(str(c))
                if hits_pnr["hits"]["total"] > 0:
                    api_results = CVRResponse(hits_pnr["hits"]["hits"][0]).get_data()
                    api_results["hit_type"] = "pnr"
                    api_results["pnr"] = int(c)
                    return api_results
                else:
                    return {"cvr": int(c), "tom_aarsag": "CVR blev ikke fundet ved opslag i cvr-registeret. Ej heller som p-nr."}

        with ThreadPoolExecutor() as executor:
            future_to_cvr = {executor.submit(process_single_cvr, c): c for c in cvr}

            for future in tqdm(as_completed(future_to_cvr), total=len(cvr)):
                result = future.result()

                if isinstance(result, list):
                    results.extend(result)
                else:
                    results.append(result)

        return pd.DataFrame.from_dict(results)    
    
    def __search_financial_based_on_cvr(self, cvr: str) -> bytes:
        self._url = self._url_regnskab
        query = {"query": {"bool": {"must": [{"term": {"cvrNummer": cvr}}, { "term": { "dokumenter.dokumentMimeType": "application" } }]}}, "size": 50}
        response = self._response(query)
        response.raise_for_status()
        return json.loads(response.content)
    
    def cvr_financial_search_to_df(self, cvr: list[int], limit: int = 1, year: int = None) -> pd.DataFrame:
        if not isinstance(cvr, list):
            raise TypeError("Input skal være en liste med cvr numre")
        print(f"Henter CVR regnskabsdata for {len(cvr)} cvr-numre")
        # Henter oversættelsesfil ind som df
        translations_df = self.__load_translations_file()
        results = []
        
        def process_single_cvr(c: int) -> dict:
            hits = self.__search_financial_based_on_cvr(str(c))
            if hits["hits"]["total"] > 0:
                off_regnskaber = self.__extract_xblr_urls(hits, year)
                if off_regnskaber:
                    results_temp = self.__handle_xblr_documents(off_regnskaber, translations_df, limit if not year else 1)
                    return results_temp
                else:
                    return {"cvr": c, "tom_aarsag": "Hits, men ingen xblr regnskaber"}
            else:
                return {"cvr": c, "tom_aarsag": "Ingen hits"}

        with ThreadPoolExecutor() as executor:
            future_to_cvr = {executor.submit(process_single_cvr, c): c for c in cvr}
            
            for future in tqdm(as_completed(future_to_cvr), total=len(cvr)):
                result = future.result()
                
                if isinstance(result, list):
                    results.extend(result)
                else:
                    results.append(result)                
                
        return pd.DataFrame.from_dict(results)
    
    def __load_translations_file(self) -> pd.DataFrame:
        translations_file = Path(Path(__file__).parent, "xbrl", "ressources", "general", "general_translations.xlsx")
        assert translations_file.exists(), f"Filen {translations_file} findes ikke."
        return pd.read_excel(translations_file)

    def __handle_xblr_documents(self, documents: list[dict], translations_df: pd.DataFrame, limit: int) -> list[dict]:
        results = []
        documents_to_proces = documents[:limit]
        for doc in documents_to_proces:
            url = doc["dokument_url"]
            downloaded_doc = self.__download_xblr_document(url)
            parsed_doc = Regnskab(downloaded_doc, doc["cvr"], doc["start_dato"], doc["slut_dato"], translations_df).get_data_as_dict()
            results.append(parsed_doc)
        return results
                
    def __download_xblr_document(self, url: str) -> bytes:
        xblr_document = requests.request(
            method="GET",
            url=url,
            headers=self._headers,
            auth=HTTPBasicAuth(self.__creds.username, self.__creds.password)
            )
        xblr_document.raise_for_status()
        return xblr_document
    
    def __extract_xblr_urls(self, hits: bytes, year: int = None) -> list[dict]:
        regnskaber = [hit for hit in hits["hits"]["hits"] if hit["_source"]["offentliggoerelsestype"]]
        regnskaber_red = []
        for r in regnskaber:
            output = {}
            output["cvr"] = r["_source"]["cvrNummer"]            
            output["start_dato"] = r["_source"]["regnskab"]["regnskabsperiode"]["startDato"]
            output["slut_dato"] = r["_source"]["regnskab"]["regnskabsperiode"]["slutDato"]
            for dok in r["_source"]["dokumenter"]:
                if dok["dokumentMimeType"] == r"application/xml":
                    output["dokument_url"] = dok["dokumentUrl"]
            # For regnskaber som ikke følger kalenderår, så følges nedenståede logik for at placere regnskabet i et aarstal
            if output["start_dato"][:4] != output["slut_dato"][:4]:
                start_of_end_year = datetime.strptime(f"{output["slut_dato"][:4]}-01-01", "%Y-%m-%d")
                diff_start = start_of_end_year - datetime.strptime(output["start_dato"], "%Y-%m-%d")
                diff_end = datetime.strptime(output["slut_dato"], "%Y-%m-%d") - start_of_end_year
                output["aarstal_afledt"] = int(output["start_dato"][:4]) if diff_start >= diff_end else int(output["slut_dato"][:4])
            else:
                output["aarstal_afledt"] = int(output["slut_dato"][:4])
            
            regnskaber_red.append(output)
            
        regnskaber_red = [dok for dok in regnskaber_red if "dokument_url" in dok]            
        regnskaber_red.sort(key=lambda x: x["start_dato"], reverse=True)
        if year:
            return [r for r in regnskaber_red if year == r["aarstal_afledt"]]
        return regnskaber_red


import configparser
from dataclasses import dataclass
from pathlib import Path


@dataclass
class ProjectConfig:
    config_file: Path

    def __post_init__(self):
        self.config = configparser.ConfigParser()
        self.config.read(self.config_file)

    @property
    def api_user(self) -> str:
        return str(self.config["CVRAPI"]["User"])

    @property
    def api_pw(self) -> str:
        return str(self.config["CVRAPI"]["Password"])

    @property
    def input_folder(self) -> Path:
        return Path(self.config["FOLDERS"]["InputFolder"])

    @property
    def output_folder(self) -> Path:
        return Path(self.config["FOLDERS"]["OutputFolder"])



from dataclasses import dataclass
from datetime import datetime
from .nace_map.nace_koder_map import nace_kode_map


@dataclass
class CVRResponse:
    """Klasse som håndteret et respons fra CVR

    OBS! Skal skrives om - alt for megen redundans omkring hvilke parametre som returneres.
    NB! Det antages at der tages udgangspunkt i det fulde response som input - altså uden selektering i output via query."""

    response: dict

    def __post_init__(self):
        if self.__invalid_input():
            raise TypeError("response er ikke en dict med det forventede format")
        self.__resp_type = self.__extract_resp_type()
        self.__meta_data = self.__extract_meta_data()

    def __invalid_input(self) -> bool:
        """Udestår"""
        return False

    def __extract_resp_type(self) -> str:
        """Udleder hvorvidt reponse er på cvr- eller p-nummer."""
        return "cvr" if self.response["_source"].get("Vrvirksomhed") else "pnr"

    def __extract_meta_data(self) -> dict:
        if self.__resp_type == "cvr":
            return self.response["_source"]["Vrvirksomhed"]["virksomhedMetadata"]
        else:
            return self.response["_source"]["VrproduktionsEnhed"][
                "produktionsEnhedMetadata"
            ]

    def get_data(self, all: bool = False, unit: str = "mdr") -> dict | list[dict]:
        self.__hoejvaekst_virk()
        if not all:
            return self.__get_latest()
        return self.__get_all(unit)

    def __get_latest(self) -> dict:
        results = self.__standard_info
        distinct_info = {
            "ant_ansatte": self.ant_ansatte,
            "ant_aarsvaerk": self.ant_aarsvaerk,
            "ant_ansatte_yyyymm": self.ant_ansatte_yyyymm,
        }
        full_results = results | distinct_info
        return full_results

    def __get_all(self, unit: str) -> list[dict]:
        results = []
        hist_results_ant_ans = self.__get_hist_data()
        if hist_results_ant_ans:
            for k in hist_results_ant_ans:
                temp_res = self.__standard_info
                distinct_info = {
                    "ant_ansatte": hist_results_ant_ans[k]["antalAnsatte"],
                    "ant_aarsvaerk": hist_results_ant_ans[k]["antalAarsvaerk"],
                    "ts_aar": hist_results_ant_ans[k]["aar"],
                    "ts_mdr": hist_results_ant_ans[k]["maaned"],
                }
                results.append(temp_res | distinct_info)
        else:
            temp_res = self.__standard_info
            results.append(temp_res)

        if unit == "mdr":
            return results
        reduced_result = self.__reduce_hist_results(results, unit)
        return reduced_result

    def __reduce_hist_results(self, complete_results: list[dict], unit: str) -> list[dict]:
        # Sikrer sortering
        ensure_sorted = sorted(complete_results, key=lambda k: (k["ts_aar"], k["ts_mdr"]), reverse=True)
        finale_res = []
        handled = []
        for k in ensure_sorted:
            if k["ts_aar"] not in handled:
                handled.append(k["ts_aar"])
                finale_res.append(k)
        return finale_res
    
    def __hoejvaekst_virk(self, year: int = None) -> int:
        """Funktion som koder hvorvidt virksomheden er en højvækst virksomhed eller ej. Definitionen følger DST's, som kan ses her: 
        https://www.dst.dk/da/Statistik/dokumentation/statistikdokumentation/hoejvaekstvirksomheder-i-danmark/indhold
        Ovenstående definerer "nye vækstvirksomheder" hvor vi kører med to - nye og eksisterende.        
        """
        # max_year: int = 2023  # Skal bare tage now() aar minus 1
        # year_list: list = [202312, 202212, 202112]
        # year_t_minus = [[year-(t*100) for t in range(0,4)] for year in year_list]
        
        # for case in year_t_minus
        pass
        

    def __get_hist_data(self) -> list[dict]:
        """Funktion som finder samtlige oplysninger om antal ansatte på tværs af måned, kvartal og årsregistreringer. Alle registreringer kodes på YYYY-MM, hvor
        tal opgjort på årsbasis placeres i måned 12, kvartal placeres i sidste måned i kvartalet.
        NB! Det er lidt forskelligt, hvad der er registreret på de forskellige typer. Antal ansatte var ex. ikke en del af kvartalstallene før 2014-07, hvorfor årstal
        prioriteres ved overlap - ex. 201312. 
        """
        hist_data = {}
        if self.__extract_mdr_data():
            mdr_data = self.__extract_mdr_data()
            hist_data = hist_data | mdr_data
        if self.__extract_aar_data():
            aar_data = self.__extract_aar_data()
            for aarmdr in aar_data:
                if int(aarmdr) not in hist_data.keys():
                    hist_data[aarmdr] = aar_data[aarmdr]
        if self.__extract_kvartal_data():
            kvartal_data = self.__extract_kvartal_data()
            for aarmdr in kvartal_data:
                if int(aarmdr) not in hist_data.keys():
                    hist_data[aarmdr] = kvartal_data[aarmdr]
        # Reducerer dict til kun at holde records og dermed ikke identifiers
        # hist_data_list = [hist_data[reg] for reg in hist_data]
        return hist_data
    
    def __extract_mdr_data(self) -> list[dict]:
        maaneddata_kodet = []
        try:
            maaneddata_kodet: list = (
                self.response["_source"]["Vrvirksomhed"]["maanedsbeskaeftigelse"]
                + self.response["_source"]["Vrvirksomhed"]["erstMaanedsbeskaeftigelse"]
            )
        except KeyError:
            maaneddata_kodet: list = self.response["_source"]["Vrvirksomhed"]["maanedsbeskaeftigelse"]
        # Opkoder data til en dict struktur
        data = {}
        if maaneddata_kodet:
            for mdr in maaneddata_kodet:
                data[int(mdr["aar"]*100+mdr["maaned"])] = {
                        "aar": mdr["aar"],
                        "maaned": mdr["maaned"],
                        "antalAarsvaerk": mdr["antalAarsvaerk"],
                        "antalAnsatte": mdr["antalAnsatte"],
                    }
        return data

    def __extract_aar_data(self) -> list[dict]:
        data = {}
        if "aarsbeskaeftigelse" in self.response["_source"]["Vrvirksomhed"]:
            for aar in self.response["_source"]["Vrvirksomhed"]["aarsbeskaeftigelse"]:
                data[(int(aar["aar"])*100)+12] = {
                        "aar": aar["aar"],
                        "maaned": 12,
                        "antalAarsvaerk": aar["antalAarsvaerk"],
                        "antalAnsatte": aar["antalAnsatte"],
                    }
        return data

    def __extract_kvartal_data(self) -> list[dict]:
        data = {}
        # Kvartalstal sammen kodes
        if "kvartalsbeskaeftigelse" in self.response["_source"]["Vrvirksomhed"]:
            for kvartal in self.response["_source"]["Vrvirksomhed"]["kvartalsbeskaeftigelse"]:
                data[int(kvartal["aar"]*100)+int(kvartal["kvartal"])*3] = {
                        "aar": kvartal["aar"],
                        "maaned": int(kvartal["kvartal"]) * 3,
                        "antalAarsvaerk": kvartal["antalAarsvaerk"],
                        "antalAnsatte": kvartal["antalAnsatte"],
                    }
        return data

    @property
    def __standard_info(self) -> dict:
        """
        Reference til de standard informationer fra CVR som bruges på tværs af forskellige funktioner.
        Lavet for at gøre det lettere at tilføje nye informationer, idet de dermed kun skal angives ét sted.
        """
        std_info = {
            "cvr": self.cvr,
            "cvr_str": str(self.cvr),
            "name": self.name,
            "kommunekode": self.kommunekode,
            "stiftelsesdato": self.stiftelses_dato,
            "h_branche": self.hbranche,
            "virkform_kode": self.virkform,
            "virkform": self.virkform_str,
            "status": self.status,
            "aktiv": self.aktiv,
            "aktiv_2": self.aktiv_2,
            "sidst_opdateret": self.sidst_opdateret
        }
        return std_info

    @property
    def cvr(self) -> int:
        if self.__resp_type == "cvr":
            return self.response["_source"]["Vrvirksomhed"]["cvrNummer"]
        else:
            try:
                return self.response["_source"]["VrproduktionsEnhed"][
                    "virksomhedsrelation"
                ][0]["cvrNummer"]
            except IndexError:
                return None

    @property
    def hbranche(self) -> int:
        if not self.__parameter_exists("nyesteHovedbranche", "branchekode"):
            return None
        return self.__meta_data["nyesteHovedbranche"]["branchekode"]
    
    @property
    def nace_grp127afd(self) -> str:
        return nace_kode_map[self.hbranche]["grp127hovedafdeling"]

    @property
    def name(self) -> str:
        if not self.__parameter_exists("nyesteNavn", "navn"):
            return None
        return self.__meta_data["nyesteNavn"]["navn"]

    @property
    def status(self) -> str:
        if not self.__parameter_exists("sammensatStatus"):
            return None
        return self.__meta_data["sammensatStatus"]

    @property
    def virkform_str(self) -> str:
        if not self.__parameter_exists("nyesteVirksomhedsform", "langBeskrivelse"):
            return None
        return self.__meta_data["nyesteVirksomhedsform"]["langBeskrivelse"]

    @property
    def virkform(self) -> int:
        if not self.__parameter_exists("nyesteVirksomhedsform", "virksomhedsformkode"):
            return None
        return self.__meta_data["nyesteVirksomhedsform"]["virksomhedsformkode"]

    @property
    def ant_ansatte(self) -> int:
        if not self.__parameter_exists("nyesteErstMaanedsbeskaeftigelse", "antalAnsatte"):
            return None
        return self.__meta_data["nyesteErstMaanedsbeskaeftigelse"]["antalAnsatte"]

    @property
    def ant_aarsvaerk(self) -> float:
        if not self.__parameter_exists("nyesteErstMaanedsbeskaeftigelse", "antalAarsvaerk"):
            return None
        return self.__meta_data["nyesteErstMaanedsbeskaeftigelse"]["antalAarsvaerk"]

    @property
    def ant_ansatte_yyyymm(self) -> int:
        if not self.__parameter_exists("nyesteErstMaanedsbeskaeftigelse", "aar"):
            return None
        if not self.__parameter_exists("nyesteErstMaanedsbeskaeftigelse", "maaned"):
            return None
        aar = self.__meta_data["nyesteErstMaanedsbeskaeftigelse"]["aar"]
        mdr = self.__meta_data["nyesteErstMaanedsbeskaeftigelse"]["maaned"]
        result = f"{(aar*100)+mdr:02d}"
        return int(result)

    @property
    def kommunekode(self) -> int:
        if not self.__parameter_exists("nyesteBeliggenhedsadresse", "kommune", "kommuneKode"):
            return None
        return int(self.__meta_data["nyesteBeliggenhedsadresse"]["kommune"]["kommuneKode"])

    @property
    def stiftelses_dato(self) -> str:
        if not self.__parameter_exists("stiftelsesDato"):
            return None
        return self.__meta_data["stiftelsesDato"]
    
    @property
    def sidst_opdateret(self) -> datetime.date:
        if self.__resp_type == "cvr":
            sidst_opdateret = self.response["_source"]["Vrvirksomhed"]["sidstOpdateret"][:10]
        else:
            sidst_opdateret = self.response["_source"]["VrproduktionsEnhed"]["sidstOpdateret"][:10]
        return datetime.strptime(sidst_opdateret, "%Y-%m-%d").date()

    @property
    def __kendte_statusudfald(self) -> dict:
        """Dict med de statuskoder, som på nuværende tidspunkt er kendte og kategoriseret i udfald på aktiv eller ej."""
        return {
            "aktiv": 1,
            "normal": 1,
            "ophørt": 0,
            "opløsteftererklæring": 0,
            "opløstefterfrivilliglikvidation": 0,
            "opløsteftergrænseoverskridendefusion": 0,
            "opløstefterfusion": 0,
            "opløstefterkonkurs": 0,
            "opløstefterspaltning": 0,
            "slettet": 0,
            "tvangsopløst": 0,
            "underfrivilliglikvidation": 0,
            "underkonkurs": 0,
            "underreassumering": 0,
            "underrekonstruktion": 0,
            "undertvangsopløsning": 0,
        }

    @property
    def aktiv(self) -> int:
        """
        Kodning af hvorvidt virksomheden er aktiv eller ej. Dette gøres på baggrund af virksomhedens status.
        Vi er i tvivl om, hvordan virksomheder som er opløste efter fusion skal betragtes. Derfor findes der et alternativ til denne
        kodning, hvor opløst efter fusion udelades fra populationen.
        """
        if self.status.lower() not in self.__kendte_statusudfald:
            print(f"Statusudfaldet: {self.status} er ikke kodet i aktiv/ikke aktiv. Skal opdateres, CVR: {self.cvr}")
        return int(self.__kendte_statusudfald[self.status.lower()])

    @property
    def aktiv_2(self) -> int:
        """Alternativ aktiv variabel som udelader opløstefterfusion fra populationen"""
        return None if self.status.lower() == "opløstefterfusion" else self.aktiv
    
    def __parameter_exists(self, *parameters) -> bool:
        """
        Hjælpe funktion til at tjekke om et parameter er i CVR response
        Parameters tager en eller flere strenge som input og returnerer True hvis parameteret er i self.__meta_data
        """
        _element = self.__meta_data
        for parameter in parameters:
            if _element is None or parameter not in _element:
                return False
            _element = _element[parameter]
        return _element is not None


# %%

from pathlib import Path
import lxml.etree as et
import pandas as pd
from dataclasses import dataclass

@dataclass
class XBRLTaxonomy:
    type: str
    
    def __post_init__(self):
        self.__validate_setup()
        
    def __validate_setup(self):
        assert self.type in ["fsa", "ifrs"], f"Den angivne taksonomi type er ikke kendt: {self.type}"
        assert self.__def_file.exists(), f"Filen {self.__def_file} findes ikke."
        assert self.__labels_file.exists(), f"Filen {self.__labels_file} findes ikke."
        
    def generate_taxo_df(self) -> pd.DataFrame:
        
        pass
        
    @property
    def __def_file(self):
        return Path(Path(__file__).parent, "ressources", self.type, f"{self.type}.xsd")
    
    @property
    def __labels_file(self):
        return Path(Path(__file__).parent, "ressources", self.type, f"{self.type}_labels.xml")

    
# %%

tester = XBRLTaxonomy("fsa")

# %%

test = Path(r"c:\Users\B375192\python_projects\ifdtools\ifdtools\cvr_api\xbrl\ressources\fsa")

# %%


taxo_def = et.parse(r"C:\Users\B375192\python_projects\ifdtools\ifdtools\cvr_api\ressources\fsa\fsa.xsd").getroot()
taxo_label = et.parse(r"C:\Users\B375192\python_projects\ifdtools\ifdtools\cvr_api\ressources\fsa\fsa-lab-da.xml").getroot()
namespaces_def: dict = taxo_def.nsmap
namespaces_label: dict = taxo_label.nsmap

# %%

# Finder relevante elementer fra definitionen af taksonomien

rel_elems = []

for elem in taxo_def.findall("*"):
    if elem.tag.endswith(r"}element"):
        temp_dict = {}
        for attr in elem.attrib:
            temp_dict[attr] = elem.get(attr)
        rel_elems.append(temp_dict)

# %%

# Finder matchende labels

rel_labels = []

# What a fucking shit show

for elem in taxo_label.findall("*"):
    if elem.tag.endswith(r"}labelLink"):
        for sub_elem in elem:
            local_name = sub_elem.tag.split(r"}")[1]
            temp_dict = {}
            if local_name == "loc":
                for attr in sub_elem.attrib:
                    if "}" in attr:
                        attr_local_name = attr.split(r"}")[1]
                        temp_dict[attr_local_name] = sub_elem.get(attr)
                        continue
                    temp_dict[attr] = sub_elem.get(attr)
                label_elem = sub_elem.getnext()
                temp_dict["label"] = label_elem.text
            if temp_dict:
                rel_labels.append(temp_dict)

# %%

taxo_defs = pd.DataFrame.from_dict(rel_elems)
taxo_labels = pd.DataFrame.from_dict(rel_labels)

# %%

# Opkoder labels til én række pr unik identifier

# Danner columns input

taxo_labels["nb"] = taxo_labels.groupby(["href"]).cumcount() + 1
taxo_labels["nb_str"] = "label_" + taxo_labels["nb"].astype(str)

# %%

taxo_labels_pivot = taxo_labels.pivot_table(
    index="href",
    columns="nb_str",
    values="label",
    aggfunc="first"
).reset_index()

# %%

# Merger de to dataframe

temp_col = taxo_labels_pivot.href.str.split(r"#", expand=True)
taxo_labels_pivot["merge_key"] = temp_col[1]
taxo_defs = taxo_defs.merge(taxo_labels_pivot, how="left", left_on="id", right_on="merge_key", suffixes=("", "_labels"))

# %%

taxo_defs.to_csv(r"c:\users\b375192\desktop\fsa_taxo_defs.csv", encoding="utf-8-sig", sep=";", decimal=",")
taxo_labels_pivot.to_csv(r"c:\users\b375192\desktop\fsa_taxo_labels.csv", encoding="utf-8-sig", sep=";", decimal=",")
# %%


test = XBRLTaxonomy("fsa")
# %%
